#include <lektor/common.h>
#include <lektor/database.h>
#include <lektor/lib/strv.h>
#include <lektor/internal/dbmacro.h>

bool
database_config_set_strv_1_2(lkt_db *db, const char *section, const struct strv key,
                             const struct strv value)
{
    static const char *SQL_STMT = "INSERT OR REPLACE INTO"
                                  " config (section, key, value)"
                                  " VALUES (?, ?, ?);\n";
    sqlite3_stmt *stmt          = 0;
    bool ret                    = false;
    int code;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_STRV(db, stmt, 2, key, error);
    SQLITE_BIND_STRV(db, stmt, 3, value, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK && code != SQLITE_DONE) {
        LOG_ERROR("DB", "Failed to insert or replace: %s", sqlite3_errmsg((sqlite3 *)db));
        goto error;
    }

    LOG_DEBUG("CONFIG", "Set [%s/" STRV_FMT "] to " STRV_FMT, section, STRV_ARG(key),
              STRV_ARG(value));
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_set(lkt_db *db, const char *section, const char *key, const char *value)
{
    static const char *SQL_STMT = "INSERT OR REPLACE INTO"
                                  " config (section, key, value)"
                                  " VALUES (?, ?, ?);\n";
    sqlite3_stmt *stmt          = 0;
    bool ret                    = false;
    int code;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_BIND_TEXT(db, stmt, 3, value, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK && code != SQLITE_DONE) {
        LOG_ERROR("DB", "Failed to insert or replace: %s", sqlite3_errmsg((sqlite3 *)db));
        goto error;
    }

    LOG_DEBUG("CONFIG", "Set [%s/%s] to %s", section, key, value);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_text(lkt_db *db, const char *section, const char *key, char *value, size_t len)
{
    static const char *SQL_STMT = "SELECT value"
                                  " FROM config"
                                  " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt          = 0;
    bool ret                    = false;
    char *row;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    row = (char *)sqlite3_column_text(stmt, 0);
    safe_strncpy(value, row, len);
    value[len - 1] = 0;
    ret            = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_text_nospace(lkt_db *db, const char *section, const char *key, char *value,
                                 size_t len)
{
    static const char *SQL_STMT = "SELECT REPLACE(value, ' ', '')"
                                  " FROM config"
                                  " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt          = 0;
    bool ret                    = false;
    char *row;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    row = (char *)sqlite3_column_text(stmt, 0);
    strncpy(value, row, len);
    value[len - 1] = 0;
    ret            = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_exists(lkt_db *db, const char *section, const char *key)
{
    static const char *SQL_STMT = "SELECT value"
                                  " FROM config"
                                  " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt          = 0;
    bool ret                    = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_int(lkt_db *db, const char *section, const char *key, int *value)
{
    static const char *SQL_STMT = "SELECT CAST(value AS INTEGER)"
                                  " FROM config"
                                  " WHERE section COLLATE nocase = ? AND key COLLATE nocase = ?;";
    sqlite3_stmt *stmt          = 0;
    bool ret                    = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    *value = sqlite3_column_int(stmt, 0);
    ret    = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_octal(lkt_db *db, const char *section, const char *key, unsigned int *value)
{
    char str_value[LKT_LINE_MAX];
    unless (database_config_get_text_nospace(db, section, key, str_value, LKT_LINE_MAX))
        return false;

    const long octal = strtol(str_value, NULL, 8);
    if (octal < 0 || octal > UINT_MAX) {
        LOG_ERROR("DB", "The octal value is negative or to big for a uint!");
        return false;
    }

    *value = (unsigned int)octal;
    return true;
}

bool
database_config_queue(lkt_db *db, const char *option, int value)
{
    static const char *SQL_STMT_TMP = "UPDATE queue_state SET %s = ? WHERE id = 42;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    sqlite3_stmt *stmt = 0;
    bool ret           = false;

    if (value < 0)
        value = 0;
    if (value > 100)
        value = 100;

    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TMP, option);
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, value, error);
    SQLITE_STEP_OK(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

void
database_config_set_optional_fields(lkt_db *db)
{
    const char *section = NULL;
#define section(_sct) section = _sct;
#define value(name, value)
#define value_opt(name, value)                                                        \
    if ((section != NULL) && !database_config_exists(db, section, name)) {            \
        FAIL_UNLESS(database_config_set(db, section, name, value),                    \
                    "Failed to set default value for option '%s/%s'", section, name); \
        LOG_WARN("CONFIG", "Enforce the option '%s/%s' <- %s", section, name, value); \
    }

#include <lektor/internal/config.def>
#undef section
#undef value
#undef value_opt
}

bool
database_validate_conf(lkt_db *db)
{
    const char *section;
#define section(_sct) section = _sct;
#define value_opt(name, value)
#define value(name, value)                                                                         \
    if (!database_config_exists(db, section, name)) {                                              \
        LOG_ERROR("CONFIG", "Missing option \"" name "\" in section \"%s\"", section);             \
        return false;                                                                              \
    } else if (STR_MATCH("externals", section)) {                                                  \
        char file[PATH_MAX];                                                                       \
        if (!database_config_get_text(db, section, name, file, PATH_MAX)) {                        \
            if (access(file, F_OK) != 0) {                                                         \
                LOG_ERROR("CONFIG", "File %s don't exists, invalid [%s/%s]", file, section, name); \
                return false;                                                                      \
            }                                                                                      \
        }                                                                                          \
    }
#include <lektor/internal/config.def>
#undef section
#undef value
#undef value_opt
    return true;
}

bool
database_config_dump(lkt_db *db, FILE *f)
{
    static const char *SQL = "SELECT section, key, value FROM config;";
    sqlite3_stmt *stmt     = 0;
    bool ret_code          = false;

    SQLITE_PREPARE(db, stmt, SQL, error);

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        if (sqlite3_column_type(stmt, 0) == SQLITE_NULL) {
            fprintf(f, "NO SECTION => %s => %s\n", sqlite3_column_chars(stmt, 1),
                    sqlite3_column_chars(stmt, 2));
        } else {
            fprintf(f, "%s => %s => %s\n", sqlite3_column_chars(stmt, 0),
                    sqlite3_column_chars(stmt, 1), sqlite3_column_chars(stmt, 2));
        }
    }

    ret_code = true;
error:
    sqlite3_finalize(stmt);
    return ret_code;
}

bool
database_config_obfuscation_check(lkt_db *db, int obfuscation)
{
    static const char *GET_SQL = "SELECT obfuscation FROM misc WHERE id = 42;";
    static const char *SET_SQL = "UPDATE misc SET obfuscation = ? WHERE id = 42;";
    sqlite3_stmt *stmt         = 0;
    bool ret_code              = false;

    SQLITE_PREPARE(db, stmt, GET_SQL, error);

    if (SQLITE_ROW == sqlite3_step(stmt)) {
        /* Obfuscation mode set */
        ret_code = (obfuscation == sqlite3_column_int(stmt, 0));
    }

    else {
        /* No obfuscation mode set */
        sqlite3_finalize(stmt);
        SQLITE_PREPARE(db, stmt, SET_SQL, error);
        SQLITE_BIND_INT(db, stmt, 1, obfuscation, error);
        SQLITE_STEP_DONE(db, stmt, error);
        sqlite3_finalize(stmt);
        ret_code = true;
    }

    ret_code = true;
error:
    sqlite3_finalize(stmt);
    return ret_code;
}
