#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import json
import shutil
import argparse

from urllib.error import URLError
from urllib.parse import urlencode
from urllib.request import urlopen
from textwrap import wrap


class CustomFormatter(
    argparse.RawTextHelpFormatter,
    argparse.RawDescriptionHelpFormatter,
):
    pass


def create_parser():
    "Creates and returns a parser for the CLI arguments"
    usage = "kagari [-h] (dl [dl options] | search [search options])"

    desc = """Client to interact with the Kurisu API. Two main modes:
    -search: searches the database with filters for info on the files
    -dl: downloads the target file with an ID, or multiple files with filters
         To download the entire database, use neither ID nor filters"""

    parser = argparse.ArgumentParser(
        usage=usage,
        description=desc,
        formatter_class=CustomFormatter,
        add_help=False,
    )

    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.\n ",
    )

    parser.add_argument(
        "mode",
        help='The mode used. Can be either "dl" or "search".',
    )

    parser.add_argument(
        "-i",
        "--id",
        help='In "dl" mode, if specified, downloads target video file.'
        + "\nIf absent in this mode, filter arguments are used instead."
        + '\nIn "search" mode, returns info of the target file.\n ',
    )

    parser.add_argument(
        "-p",
        "--path",
        help='In "dl" mode, indicates the folder in which to store the file.'
        + "\nIf absent in this mode, defaults to current working directory."
        + '\nIn "search" mode, if specified, saves the output as a file'
        + " at the given path.\n ",
    )

    parser.add_argument(
        "-C",
        "--check",
        action="store_true",
        help='In "dl" mode, first displays info on the targeted'
        + "\nfile(s), and asks for confirmation before downloading.\n ",
    )

    parser.add_argument(
        "-a",
        "--author",
        help='In either mode, filters by author. (case insensitive)',
    )

    parser.add_argument(
        "-c",
        "--category",
        dest="cat",
        help='In either mode, filters by category.\n ',
    )

    parser.add_argument(
        "-t", "--type", help='In either mode, filters by type.\n '
    )

    parser.add_argument(
        "-s",
        "--string",
        dest="search",
        help='In either mode, filters on the names of sources and songs'
        + "\nmatching the given string. (case insensitive)\n ",
    )

    parser.add_argument(
        "-n",
        "--newer-than",
        dest="newer_than",
        help='In either mode, filters on the date of uploading,'
        + "\nreturning files after the date (format YYYY-MM-DD)\n "
    )

    parser.add_argument(
        "-r",
        "--random",
        action="store_true",
        help='In "search" mode, randomizes the order of the files '
        + "in the response.\n ",
    )

    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help='In "search" mode, output the result as a json.',
    )

    return parser


def make_request(url):
    "Executes an HTTP request with basic error handling"
    try:
        response = urlopen(url)
    except URLError as err:
        if hasattr(err, "reason"):
            emsg = f"Failed to reach target URL. Reason: {err.reason}"
            print(emsg)
            sys.exit(1)
        elif hasattr(err, "code"):
            emsg = f"The server could not fulfill the request. Code {err.code}"
            print(emsg)
            sys.exit(1)
    return response


def search_query(args):
    "Searches the database with ID or filters, and returns the response json"
    url = "https://kurisu.iiens.net/api"
    url_values = dict()

    # If ID, it's the only parameter needed
    if args.id:
        url_values["id"] = args.id
    # No ID, building the query parameters
    else:
        arg_list = ["author", "cat", "type", "search", "random", "newer_than"]
        for arg_name in arg_list:
            arg = getattr(args, arg_name)
            if arg:
                url_values[arg_name] = arg

    url_args = urlencode(url_values)
    if url_args:
        url += "?" + url_args

    response = make_request(url)

    try:
        kara_json = json.loads(response.read().decode("utf-8"))
    except ValueError:
        emsg = "The response JSON could not be loaded."
        print(emsg)
        sys.exit(1)

    # Handling an error json
    if "error" in kara_json:
        default_emsg = "An unknown error happened when searching the database."
        emsg = kara_json.get("message", default_emsg)
        print(emsg)
        sys.exit(1)

    return kara_json


def build_kara_name(data):
    "Builds the name of the file from the response dictionary"
    kara_id = data["id"]
    src_name = data["source_name"]
    song_name = data["song_name"]
    song_type = data["song_type"]
    song_nbr = data["song_number"]
    category = data["category"]
    language = data["language"]
    author = data["author_name"]

    kara_name = f"{kara_id} - {src_name} - {song_type}{song_nbr} - {song_name}"
    kara_additional = f" ({category}/{language}) by {author}"

    return kara_name, kara_additional


def kara_prompt(kara_string="", kara_nbr=0):
    "Simple yes/no prompt to confirm the download or not"
    prompt_str = "You are about to download "
    if kara_string:
        prompt_str += kara_string
    else:
        prompt_str += f"{kara_nbr} files"
    prompt_str += "\nDo you want to proceed ? ([y]/n) "

    answers = {"y": True, "n": False, "ye": True, "yes": True, "no": False}
    ans = None
    while ans not in answers.keys():
        try:
            ans = input(prompt_str).lower()
            if not ans:
                ans = "y"
        except KeyboardInterrupt:
            ans = "n"
            print("")
        except EOFError:
            ans = "y"

    return answers[ans]


def dl_mode(args, kara_json):
    "Downloads the targeted files"

    kara_id = None
    kara_string = ""
    kara_nbr = 0

    if type(kara_json) is not list:
        kara_name, kara_additional = build_kara_name(kara_json)
        kara_string = '"' + kara_name + '"' + kara_additional
        kara_id = kara_json["id"]
    else:
        kara_nbr = len(kara_json)

    # If asked for check, display the prompt
    if args.check:
        prompt = kara_prompt(kara_string, kara_nbr)
        if not prompt:
            print("Download cancelled.")
            return

    if kara_id:
        url = f"https://kurisu.iiens.net/api/download/{kara_id}"
        path = args.path if args.path else os.getcwd()
        filename = path + "/" + kara_name + ".mkv"

        print(f'Downloading and writing video as "{filename}"')
        response = make_request(url)

        # Writing the video file
        with open(filename, "wb") as out_file:
            shutil.copyfileobj(response, out_file)
    else:
        print(f"Download in progress: 0/{kara_nbr}", end="")
        sys.stdout.flush()
        for curr_nbr, kara in enumerate(kara_json):
            kara_id = kara["id"]
            kara_name, _ = build_kara_name(kara)
            url = f"https://kurisu.iiens.net/api/download/{kara_id}"
            path = args.path if args.path else os.getcwd()
            filename = path + "/" + kara_name + ".mkv"

            response = make_request(url)

            # Writing the video file
            with open(filename, "wb") as out_file:
                shutil.copyfileobj(response, out_file)
            print(f"\rDownload in progress: {curr_nbr + 1}/{kara_nbr}", end="")
            sys.stdout.flush()
        print("")
    print("All done !")


def search_mode(args, kara_json):
    "Performs a search request on the database and returns the answer"
    if args.json:
        pretty_output = json.dumps(kara_json, indent=4, ensure_ascii=False)
        pretty_output += "\n"
    else:
        pretty_output = ""
        for res in kara_json:
            new_kara = " NEW!" if res["is_new"] == "1" else ""
            output_info = build_kara_name(res)
            output_string = (
                f"[{res['id']}] {output_info[0]}\n"
                + f"{output_info[1]} in {res['author_year']}{new_kara}\n"
                + f" Size: {res['size']} Kio\n"
            )
            pretty_output += output_string
            if res["popularity"]:
                pretty_output += " Popularity: " + res["popularity"] + "\n"
            if res["upload_comment"]:
                for comment_line in wrap(
                    res["upload_comment"], fix_sentence_endings=True
                ):
                    pretty_output += " " + comment_line + "\n"
            pretty_output += "\n"

    # Displaying the received JSON, or writing it to a file
    if not args.path:
        print(pretty_output, end="")
    else:
        try:
            with open(args.path, "w") as output_file:
                output_file.write(pretty_output)
        except OSError as err:
            emsg = "Cannot create/open the search output file: "
            emsg += str(err)
            print(emsg)
            sys.exit(1)


def run():
    "Runs the command"
    parser = create_parser()
    args = parser.parse_args()

    kara_json = search_query(args)

    if args.mode == "dl":
        try:
            dl_mode(args, kara_json)
        except KeyboardInterrupt:
            print("\nDownload interrupted by user input.")
    elif args.mode == "search":
        search_mode(args, kara_json)
    else:
        print(f'Unknown mode "{args.mode}". Use either "dl" or "search".')
        sys.exit(1)


if __name__ == "__main__":
    run()
